# Copyright (c) 2017 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


# Attempt to import the requisite Python modules.
# Print a descriptive error message to the user
# if any of these dependencies cannot be resolved.
#
# collections.Callable works for Python2 and all Python version 3 variants.
#
import os
import sys
import shlex
import select
import subprocess
import time
import fcntl
import socket
import six
import re
from collections import Callable

HTML_LOOKUPS = {
    "Log In": "_eventId_submit",
    "Product URL": "product-url",
    "Program": "program_chosen",
    "Red Hat Product": "rh_product_chosen",
    "Version": "rh_product_version_chosen",
    "Architecture": "arch_chosen",
    "Format": "format_chosen",
    "Build": "build",
    "Product Name": "product_option_chosen",
    "Vendor Name": "vendor_chosen",
    "Component": "component_chosen",
    "Submit": "button",
    "Continue": "button"
}

# Global to save the question being asked to match later
questionInKey = None

# This helper function checks if a /dev/<disk> is mounted and
# then returns True or False.
#
# Ideal Condition for isDiskMounted function is FALSE.
def isDiskMounted(diskDevice):
    isMountedCommand = "grep -w -o " + diskDevice + " /proc/mounts"
    mounted = runShellCommand(isMountedCommand)[1]
    mounted = mounted.decode().strip() if isinstance(mounted, bytes) else mounted.strip()
    if diskDevice in mounted:
        return True
    return False

## This helper function checks if a /dev/<disk> is a boot partition ("/boot")
## and then returns True or False.
##
## Ideal Condition for isDiskBootPart function is FALSE.
def isDiskBootPart(diskDevice):
    isBootCommand = 'lsblk -i --noheadings --paths --output=NAME,MOUNTPOINT ' + diskDevice
    bootDevice = runShellCommand(isBootCommand)[1]
    bootDevice = bootDevice.decode().strip() if isinstance(bootDevice, bytes) else bootDevice.strip()

    # Extra space is important for pattern match.
    diskDevicePattern = '^{0} '.format(diskDevice)
    bootPartPattern = '/boot$'
    patterns = [ diskDevicePattern , bootPartPattern]

    # NOTE: re.search executes twice, once for each pattern.
    #  So both MUST be true for this to work.
    for pattern in patterns:
        if not re.search(pattern, bootDevice):
            return False
        return True
#
## This helper function checks if a /dev/<disk> is of TYPE disk from
## command lsblk --noheadings --raw --paths --output=NAME,TYPE /dev/sda
##
## Ideal Condition for isDiskType function is FALSE
## False implies that what is displayed/reported from TYPE column is "lvm" or "part"
## True implies device from the TYPE column TYPE "disk" - Not an Ideal choice
#def isDiskType(diskDevice):
#    isDiskType = 'lsblk --noheadings --raw --paths --output=NAME,TYPE ' + diskDevice
#    deviceType = runShellCommand(isDiskType)[1].strip()
#    diskDevicePattern = '^{0} disk'.format(diskDevice)
#    patterns = [ diskDevicePattern ]
#
#    # NOTE: re.search executes twice, once for each pattern.
#    #  So both patterns MUST be true for this to work.
#    for pattern in patterns:
#        if re.search(pattern, deviceType):
#            return True
#        return False
#
## This helper function is only to check if a /dev/<disk> has SWAP on it
## and then returns True or False.
##
## Ideal Condition for isDiskSwap function is TRUE.
##
## NOTE: Later on, you can still test a disk that does NOT have SWAP, this is just
##       the Ideal condition.
def isDiskSwap(diskDevice):
    isSwapCommand = 'lsblk --output=NAME,MOUNTPOINT ' + diskDevice
    swapDevice = runShellCommand(isSwapCommand)[1].split()
    if '[SWAP]' in swapDevice:
        return True
    return False

## This function will loop through all the presented choices for any of
## Storage tests and will return the ideal choice as the response.
## which will be written as the choice you wanted to the presented
## question of which disk to test?
##
## Your kilometers may vary on what "Ideal" means to you.
## Currently for a disk to end up in the idealChoice set the
## Ideal conditions in order is:
##
##                  Ideal
##                  Boolean
## Function         Returned
## -------------------------
## isDiskType       False
## isDiskMounted    False
## isDiskBootPart   False
## isDiskSwap       True
##
##
## Once a disk has been added to the notIdealSet it does NOT ever get
## removed from it.
##
# Assuming the choices of Disks has already been vetted by
# the certification test(s) at this point, no need to prove the disks exist.
#
# NOTE: Even at the end of the analysis we may still have NO Ideal disk(s).
#       So we will just return the original default response.
#
# Feel free to add more functions to zero in on the Ideal choice.
def selectDiskToTest(choiceTuple, default):
    # Initialize Set
    notIdealSet = set()
    idealChoiceSet = set()

    for diskItem in choiceTuple:
        diskDevice = '/dev/' + diskItem.strip()

        # Begin decision procress
        # Is Disk is Mounted?
        #    True  = NOT IDEAL Case
        #    False = IDEAL Case
        if isDiskMounted(diskDevice) is True:
            # True: NOT IDEAL Case (add to notIdealSet)
            notIdealSet.add(diskItem)

        # is Disk the Boot Partition?
        #    True  = NOT Ideal Condition
        #    False = Ideal Condition
        if isDiskBootPart(diskDevice) is True:
            # True: NOT IDEAL Case (add to notIdealSet)
            notIdealSet.add(diskItem)

        # Does Disk have SWAP?
        #    True  = Ideal Condition
        #    False = NOT Ideal Condition
        if isDiskSwap(diskDevice) is False:
            # False: NOT IDEAL case (add to notIdealSet)
            notIdealSet.add(diskItem)

        # Now see if there is only one disk choice left (the Ideal Choice!) by subtracting
        # the two python sets.
        choiceSet = set(choiceTuple)
        idealChoiceSet = choiceSet - notIdealSet

    # After subtraction decide what to do.
    #
    # There is nothing in the idealChoiceSet therefore since there was no
    # ideal disk found then return the original default response.
    # NOTE: Original response already has the "\n" appended to it. So do NOT add it here.
    if len(idealChoiceSet) == 0:
        print ("WARNING: Zero choices in ideal Set therefore returning original default response\n")
        return default

    # So a little overloading here. If there
    # is only one choice, our job is done, but if multiple
    # choices exist, pick a random choice anyway.
    #
    # NOTE/Warning:
    # Because python set's have no indexing and are unordered the choice returned
    # in the case of MULTIPLE CHOICES within the idealChoiceSet the item returned
    # will be random! In adding an item to a python set the order of adding is
    # not guaranteed!
    return list(idealChoiceSet)[0]

## This function will check rdmaInterface for an UP state AND
## if it has a well-formed IPv4 address then return True if otherwise
## return False
##
## NOTE:
##   In our own internal Red Hat lab environment Infiniband devices ending with '.8006' is not to be
##   used for testing. This is strictly for RDMA development use.
#def isRdmaDeviceReady(rdmaInterface):
#    # activate skip8006 when certpx-3877 is resolved
#    # skip8006='8006'
#    # if rdmaInterface.endswith(skip8006):
#    #     return False
#    isStateUp ='UP'
#    ipv4Pattern = re.compile(r'(\d+.\d+.\d+.\d+)')
#    command = 'ip -f inet addr show ' + rdmaInterface
#    output = runShellCommand(command)[1]
#    interfaceState = (str(re.findall(r' state (.*?) group ', str(output)))).strip()
#    if isStateUp in interfaceState:
#        # check if it has a well formed IPv4 address
#        inet = (str(re.findall(r' inet (.*?)\/', str(output)))).strip('[]"\'')
#        ipv4match = ipv4Pattern.findall(inet)
#        if ipv4match:
#            return True
#    return False
#
#
## This function will take rhcert-cli print output and parse out
## check to see if the rdma_interface you passed into the function exists
## It will return all that in a tuple containing input test parameters for 
## needed to execute rhcert-cli tests. 
##
## For example the return tuple will contain:
## hwcert/rdma/Infiniband_QDR --device mlx4_0_devicePort_1_netDevice_mlx4_ib0.8002
## the test can then be executed by prepending: rhcert-cli run --test aka:
## rhcert-cli run --test hwcert/rdma/Infiniband_QDR --device mlx4_0_devicePort_1_netDevice_mlx4_ib0.8002
##
## Warning:
##   For this function you need to validate that the rdma_interface you want to
##   test from the internal Red Hat testing environment is testable. You can
##   test any interface you want but, not all of them are to be used for testing.
##
## NOTE:
##   In our own internal Red Hat lab environment Infiniband devices ending with '.8006' is not to be
##   used for testing. This is strictly for RDMA development use. These will
##   most likely fail
#def rdmaParseForSingleTestCmd(output, rdma_test, rdma_interface):
#    output = output.split('\n')
#
#    #print ("SingleTestCmd rdma_test: {0}".format(rdma_test))
#    #print ("SingleTestCmd rdma_interface: {0}".format(rdma_interface))
#
#    # attempt a validility check
#    pattern = re.compile (r'([0-9a-zA-Z_\.]+)')
#    #
#    testItemTuple = ()
#    testItemList = []
#
#    # rdma_test_valid = pattern.match(rdma_test)
#    # rdma_interface_valid = pattern.match(rdma_interface)
#    # print ("rdma_test_valid: {0}".format(rdma_test))
#    # print ("rdma_interface_valid: {0}".format(rdma_interface))
#    # if not rdma_test_valid:
#    #     print ("ERROR: rdma_test_valid: {0}".format(rdma_test))
#    #     assert False, "ERROR: rdma_test does not seem to be valid\n"
#        
#    # if not rdma_interface_valid:
#    #     print ("ERROR: rdma_interface_valid: {0}".format(rdma_interface))
#    #     assert False, "ERROR: rdma_interface does not seem to be valid\n"
#
#    # create part of the test name path
#    test_path = 'hwcert/rdma/'
#
#    for line in output:
#
#        # Do not use loopback interfaces with 127.0.0.1 (I.e. 'lo:' devices) 
#        skipLocalHost='127.0.0.1'
#        if skipLocalHost in line:
#            continue
#
#        if rdma_interface in line:
#            # get entire network device name 'mlx4_0_devicePort_1_netDevice_mlx4_ib0.8002'
#            netDevice = (str(re.findall(r' (.*?)$', str(line))).strip(' []"\''))
#            rdma_test = str((test_path + rdma_test).strip('[]()"\'\-,'))
#            rdma_device = str((netDevice).strip('[]()"\'\-,'))
#            testItemList.append(rdma_test)
#            testItemList.append(rdma_device)
#
#            # Only get the first testable RDMA Interface that matches the pattern.
#            break
#
#    if not len(testItemList):
#        # nice to see in log file if no RDMA interfaces from the feature file that do NOT exist on SUT
#        print ("WARNING: SingleTestCmd - testItemList does not contain the RDMA interface from the feature file!\n")
#
#        # Do not leave the list or tuple empty. This gets unpacked later and gets
#        # tested for values of None.
#        # rdma_test = None, rdma_device = None
#        testItemList = [(None), (None)]
#
#    # cast to read-only tuple
#    testItemTuple = tuple(testItemList)
#    return testItemTuple
#
## This function expects to be passed an RDMA string provided from the question being
## asked by the test. Using the provided string determine and return only the
## numbered choice that it best matches with.
#def selectLTSNetworkInterface(providedLTSNetworkDevice, output, default):
#    response = default
#
#    # providedLTSNetworkDevice COULD be one of these devices:
#    #  Infiniband = mlx4_0_devicePort_1_netDevice_mlx4_ib0.8002
#    #             = mlx4_0_devicePort_1_netDevice_mlx4_ib0
#    #
#    #  RoCE       = mlx5_0_devicePort_1_netDevice_mlx5_roce
#    #
#    #  iWarp      = cxgb4_0_devicePort_1_netDevice_cxgb4_iw
#    #
#    #  Omnipath   = hfi1_0_devicePort_1_netDevice_hfi1_opa0
#    #             = hfi1_0_devicePort_1_netDevice_hfi1_opa0.8022
#    infinibandType = '_ib'
#    iwarpType = '_iw'
#    roceType = '_roce'
#    omnipathType = '_opa'
#
#    # Initialization
#    rdmaQuestion = 'Which ip address (LTS) would you like to test'
#    matchEnterQuestion = 'Enter the choice number or the choice value'
#    sutNetworkDevice = (str(re.findall(r'netDevice_(.*?)$', str(providedLTSNetworkDevice)))).strip('[]"\'')
#
#    # pattern = Find the provided decimal choice from output and return it as 
#    # the response to the question
#    pattern = re.compile(r'(^\d+).')
#    # Boolean to start looking at the output very close to what your looking for
#    lineStart = False
#
#    for line in output:
#        # one shot- Once you see the question continue. 
#        #if questionInKey in line:
#        if rdmaQuestion in line and lineStart is False:
#            # want 'mlx4_ib0.8002'
#            sutNetworkDevice = (str(re.findall(r'netDevice_(.*?)$', str(providedLTSNetworkDevice)))).strip('[]"\'')
#            # print ("selectLTSNetworkInterface sutNetworkDevice: {0}".format(sutNetworkDevice))
#            # Once seen, set True
#            lineStart = True
#            continue
#            
#        matchPattern = str(pattern.match(line))
#        if lineStart and matchPattern and sutNetworkDevice in line:
#            response = str(pattern.findall(line))
#            # print ("selectLTSNetworkInterface Infiniband response: {0}".format(response))
#            # Return immediately 
#            return str(response).strip('[]"\'')
#
#        if matchEnterQuestion in line:
#            # You hit the end, so no sense in looping any more
#            break
#
#    # No match was found return default
#    return str(response).strip('[]"\'')
#
## This function is responsible for printing a helpful error message
## when an exception <e> is thrown. It will reraise the exception if
## <raiseError> is true, causing the step to fail.
def printError(e, raiseError=False):
    print("{0}: {1}".format(e.__class__, str(e)))
    if raiseError:
        raise e
#
## This function will parse the choices presented by the Certification test from stdout
## stuffed into variable output.
## choices == disk choices presented by any Storage test. (I.e. sda, sda1, sda2, all)
##
## The goal is to make a better choice then replace the default response, only
## if necessary.
##
## Once the stdout has processed place all the choices into a read only
## tuple. Then pass this tuple to each function as necessary
## to perform some analysis and then bubble up the "desired" or Ideal disk choice as the
## response through this function so the choice can be written to the question being asked
## by the certification test.
##
## output  = Is the stdout buffer, so the presented choices can be extracted and acted upon.
## default = saved default response that was originally going to be used.
## *args   = Is any other arguments that could be passed.
## choiceDiskTuple = Contains the initial read only disk choices presented by the certification test
def answerDiskController(output, default, *args):

    # remove "\n" here, so we can consistently handle any response later
    default = default.strip()

    # Parse output for the given choices and store for
    # processing
    pattern = re.compile(r'^\d+. (\w+)')
    choiceDiskTuple = ()

    output = output.split('\n')

    # Use the list to initially get the choices but, throw it
    # away in favor of a tuple.
    choiceList = []

    for line in output:
        match = pattern.findall(line)
        if match:
            choiceList.append(str(match).strip('[]"\''))

    # Cast from a list to a Tuple so to force to be read only
    choiceDiskTuple = tuple(choiceList)

    # Now begin the process of selecting the "ideal" choice for disk to test.
    response = selectDiskToTest(choiceDiskTuple, default)

    # Now Do NOT forget to append the newline "\n" back in here to the response otherwise
    # the test will hang forever waiting for that newline to move on and begin the test.
    return response + '\n'

## This function will parse the choices presented by the Certification test from stdout.
## The network tests are designed that the LTS/SUT each have a matching configured interface
## of the same genre. 
## The test will provide an interface to use on the LTS. All you need to do
## is match it to the choices being presented.
#
##  For example:
# # Which ip address (LTS) would you like to test mlx4_0_devicePort_1_netDevice_mlx4_ib0.8002 with:
#
## The choices presented to you on the output will look similiar to:
##  1. inet 127.0.0.1/8 scope host lo
##  2. inet 10.16.41.240/21 brd 10.16.47.255 scope global noprefixroute dynamic lom_1
##  3. inet 172.31.0.10/24 brd 172.31.0.255 scope global noprefixroute dynamic mlx4_ib0
##  4. inet 172.31.2.10/24 brd 172.31.2.255 scope global noprefixroute dynamic mlx4_ib0.8002
##  5. inet 172.31.6.10/24 brd 172.31.6.255 scope global noprefixroute dynamic mlx4_ib0.8006
#
## NOTE: In this case the correct match to mlx4_0_devicePort_1_netDevice_mlx4_ib0.8002 is
## choice "4".
#
## output  = Is the stdout buffer, so the presented choices can be extracted and acted upon.
## default = saved default response that was originally going to be used.
## *args   = Is any other arguments that could be passed.
#def answerNetworkInterface(output, default, *args):
#    # remove "\n" here, so we can consistently handle any response later
#    default.strip()
#    output.split('\n')
#
#    response = default
#    respPattern = re.compile (r'^\d+')
#
#    # Small sample of provided LTS network interface device of what the 
#    # string COULD possibly look like: 
#    #    Infiniband = mlx4_0_devicePort_1_netDevice_mlx4_ib0.8002
#    #    RoCE       = mlx5_0_devicePort_1_netDevice_mlx5_roce
#    #    iWarp      = cxgb4_0_devicePort_1_netDevice_cxgb4_iw
#    #    Omnipath   = hfi1_0_devicePort_1_netDevice_hfi1_opa0.8022 
#    #
#    # providedLTSNetworkDevice ='mlx4_0_devicePort_1_netDevice_mlx4_ib0.8002'
#    providedLTSNetworkDevice = (str(re.findall(r'Which ip address \(LTS\) would you like to test (.*?) with:', str(output))))
#    response = str(selectLTSNetworkInterface(providedLTSNetworkDevice, output, default))
#    matchResponse = respPattern.findall(response)
#    if matchResponse:
#        return response + '\n'
#    else:
#        # This is bad. If no decimal value then the logic to find it is flawed
#        assert False, "ERROR: Expected a decimal value in response. Stop here!"

# This function is responsible for running a shell command (<command>),
# and returning its stdout, stderr, and exit status.
# For rhcert-related commands, this function is also responsible
# for detecting when the process has prompted for input, and providing
# a reasonable response to prevent the process from blocking on input.
def runShellCommand(command, rhcertCommand=False):
    shellCommand = command
    child = None
    #rhVer = os.popen('cat /etc/redhat-release').read().split()[-2].split('.')[0]
    rhVer = os.popen('cat /etc/redhat-release').read().split('release')[1].split()[0].split('.')[0]
    pyVer = sys.version.split()[0].split('.')[0]
    usePY3 = False
    if pyVer == '3' and rhVer == '8':
        usePY3 = True

    # redhat-certification scripts for which this function
    # will be responsible for providing input (if necessary)
    rhcertExecutables = ["rhcert-cli", "rhcertd", "rhcert-ci", "rhcert-util"]

    #stdout = ""
    #stderr = ""
    
    child_output  = { 
        "stdout": "", 
        "stderr": "",
    }

    # Data structure that details known prompts for the hwcert info test,
    # along with some instruction for how to satisfy them.
    # If a key of rhcertPrompts exists within an output string,
    # and all prerequisite strings (if specified) have been
    # previously printed within the output stream, then
    # provide the associated response.

    rhcertPrompts = {
        "lock file /var/lock/subsys/rhcert found": {
            "prerequisites": list(),
            "response": "yes\n",
        },
        "/var/rhcert/results.xml": {
            "prerequisites": list(),
            "response": "yes\n",
        },
        "Would you like to continue? (yes|no) ": {
            "prerequisites": [
                "Warning: The hardware devices have changed.",
            ],
            "response": "yes\n",
        },
        "Would you like to install them now? (yes|no) ": {
            "prerequisites": [
                "The following packages are required for testing:",
            ],
            "response": "yes\n",
        },
        "Would you like to re-try installing the missing packages?": {
            "prerequisites": [
                "The following packages are required for testing:"
            ],
            "response": "no\n",
        },
        "suspend?": {
            "prerequisites": [
                "This test will suspend the operating system.",
                "Please resume by pressing the power button after suspend has completed.",
            ],
            "response": "no\n",
        },
        "Does this system have a function key (Fn) to suspend the system to mem?": {
            "prerequisites": [
                "This test will suspend the operating system.",
                "Please resume by pressing the power button after suspend has completed.",
            ],
            "response": "no\n",
        },
        "Does this system have a function key (Fn) to suspend the system to disk?": {
            "prerequisites": [
                "This test will suspend the operating system.",
                "Please resume by pressing the power button after suspend has completed.",
            ],
            "response": "no\n",
        },
        "Please describe the need for these repos": {
            "prerequisites": list(),
            "response": "redhat-certification-ci\n",
        },
        "Do you want to proceed with the operation?": {
            "prerequisites": [
                "This host is not registered with any server.",
            ],
            "response": "yes\n",
        },
        "which disk would you like to test": {
            "prerequisites": list(),
            "response": lambda output : answerDiskController(output, "all\n")
        },
        "Which ip address (LTS) would you like to test": {
            "prerequisites": list(),
            "response": lambda output : answerNetworkInterface(output, "1\n"),
        },
        "Planning the fv_storage test. Do you want to assign a device of the SUT to the guest machine for testing the storage in the guest machine?": {
            "prerequisites": list(),
            "response": "No\n",
        },
        "Do you want to register the system? (yes|no) ": {
            "prerequisites": [
                "This test system is not registered with any server(LTS).\n",
            ],
            "response": "no\n",
        },
        "Enter 'confirm' if you are really sure you want to do this: ": {
            "prerequisites": [ 
                "This command will delete all data stored locally, including test results and test systems \n",
            ],
            "response": "confirm\n",
        },
        "Are you sure you want to delete all current test results? (yes|no) ": {
            "prerequisites": list(),
            "response": "yes\n",
        },
        "Also remove certification data? (yes|no) ": {
            "prerequisites": list(),
            "response": "yes\n",
        },
        "Do you wish to continue? (yes|no) ": {
            "prerequisites": [
                "Warning: Test runs in debug mode are invalid for certification.\n",
                "The following tests will be skipped: self_check, sosreport\n",
            ],
            "response": "yes\n",
        },
        "Would you like to continue? (yes|no) ": {
            "prerequisites": list(),
            "response": "yes\n",
	},
        "Would you like to save the results to a local test server first? (yes|no) ": {
            "prerequisites": list(),
            "response": "no\n",
	},
        "Is this test for a specific device? (yes|no) ": {
            "prerequisites": list(),
            "response": "no\n",
	},
        "Verification failed, would you like to continue testing? (yes|no) ": {
            "prerequisites": list(),
            "response": "yes\n",
	}
    }

    # If the shell command was provided to this function as a string,
    # use the shell's lexical parser to tokenize the input, and store
    # the tokens within a list.
    #if isinstance(command, (str, unicode)):
    if isinstance(command, six.string_types):
        try:
            shellCommand = shlex.split(command)
        # Catching base class 'Exception' here because shlex does not
        # document which exception(s) are thrown for shlex#split.
        except Exception as e:
            #printError(e, raiseError=True)
            print(e)

    # If __runShellCommand was called to run a script that is provided by
    # redhat-certification (i.e., rhcert-cli, rhcert-ci, rhcertd, etc.),
    # We need to modify the call in order to properly perform I/O
    # operations with the child process. Specifically, we need to:
    # (1) Invoke the script via python explicitly, so that the '-u' option
    #     can be used. The -u option forces the child process to not
    #     buffer its stdout and stderr streams, which is necessary to
    #     detect when the process is waiting for input via stdin.
    # (2) Provide the absolute path to the specified redhat-certification
    #     script. We need to specify the absolute path to the script
    #     because we are running it explicitly via Python. This means that
    #     the script's location within the filesystem will not be resolved
    #     automatically via the PATH environment variable when the shell
    #     command is evaluated.
    for executable in rhcertExecutables:
        if executable in shellCommand[0]:
            fullPath = runShellCommand(
                "which {0}".format(shellCommand[0])
            )[1].strip()
            if usePY3:
                shellCommand = ["python3", "-u", fullPath] + shellCommand[1:]
            else:
                shellCommand = ["python", "-u", fullPath] + shellCommand[1:]
    try:
        # Create the child process to run <command>.
        child = subprocess.Popen(
            shellCommand,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # Allow the child process time to properly initialize.
        time.sleep(1)

        # If we're running an rhcert-cli command, we need to monitor
        # the process's stdout to determine if the process requires any
        # information from the test via stdin. We'll need to carefully
        # manage any I/O with the child process to ensure it does not hang.
        if rhcertCommand:
            # Set the pipe ends (file descriptors) connected to the child
            # process to non-blocking mode. This will force the parent to
            # immediately read any output the child has printed, even
            # if a full line of output is unavailable.
            # This is necessary for us to identify when the child
            # process is blocked waiting for input.
            fcntl.fcntl(child.stdout.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)
            fcntl.fcntl(child.stderr.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)

            # Default values for the exit status, stdout logging,
            # and stderr logging of the rhcert child process that will
            # be returned to the calling function.
            childExitStatus = None

            # Create a select() polling object. This works similarly to the
            # Unix select() system call, and will help us determine when the
            # child process has new output to process.
            selectInterface = select.poll()

            # Register the child process's stdout and stderr for output.
            selectInterface.register(child.stdout, select.POLLIN)
            selectInterface.register(child.stderr, select.POLLIN)

            # Begin polling the child process for output.
            def promptRHCert(prompt=None):
                pollEvents = 0
                maxPollRetry = 5000
                
                while True:
                    if prompt and pollEvents >= maxPollRetry:
                        if isinstance(rhcertPrompts[prompt]["response"], Callable):
                            # send the output from stdout so it can be processed
                            processedResponse = str(rhcertPrompts[prompt]["response"](child_output['stdout']))
                            if usePY3:
                                child.stdin.write(processedResponse.encode('utf-8'))
                                child.stdin.flush()
                            else:
                                child.stdin.write(processedResponse)
                        else:
                            if usePY3:
                                child.stdin.write(rhcertPrompts[prompt]["response"].encode('utf-8'))
                                child.stdin.flush()
                            else:
                                child.stdin.write(rhcertPrompts[prompt]["response"])

                        return promptRHCert()

                    pollData = selectInterface.poll(1)

                    # This list will store the channels whose select event
                    # has a hang-up status. The child process has terminated
                    # if its stdout and stderr channels enter the "POLLHUP"
                    # state. 
                    channelHUP = list()

                    # Nice Debug for polling events
                    #if prompt:
                    #    if not pollEvents % 10000:
                    #        print("pollData: {0}".format(pollData))
                    #        print("pollEvents: {0}".format(pollEvents))
                    #        print("prompt: {0}".format(prompt))

                    if len(pollData) == 0:
                        pollEvents +=1
                        continue

                    pollEvents = 0

                    # Iterate through the channels that were polled to examine
                    # any I/O events that may have occurred.
                    for selectEvent in pollData:
                        # Check to see if the child has written any new data
                        # to stdout. The business logic here is a bit
                        # more complicated, because we need to determine
                        # whether or not the child process requires
                        # any additional information.
                        if selectEvent[0] == child.stdout.fileno() and selectEvent[1] == select.POLLIN:
                            # child.stdout.flush()
                            # Perform a non-blocking read on the child's stdout
                            # channel.
                            output = child.stdout.read()
                            # Log the output that was retrieved from the child
                            # process, and append it to our cumulative output
                            # string.
                            print(output.strip())
                            if usePY3:
                                child_output['stdout'] = child_output['stdout'] + output.decode()
                            else:
                                child_output['stdout'] = child_output['stdout'] + output
                            
                            # Determine if the output from the child process
                            # contains any strings that are known to require
                            # user input. These strings are detailed by the
                            # top-level keys in the rhcertPrompts data structure.
                            for key in rhcertPrompts.keys():
                                if usePY3:
                                    op = output.decode()
                                else:
                                    op = output
                                #if key in output.decode():
                                if key in op:
                                    # If key is in output, then save it for answering the question later,
                                    questionInKey = rhcertPrompts[key]

                                    # data structure, and ensure that each of
                                    # the contextual strings in the object's
                                    # prerequisites list have also been found
                                    # within the child's output stream.
                                    # If a prompt and all of its contextual
                                    # prerequisites have been found within the
                                    # child's output stream, write the
                                    # pre-defined response to the child process
                                    # via stdin.
                                    prerequisitesSatisfied = True
                                    for prereq in rhcertPrompts[key]["prerequisites"]:
                                        if prereq not in child_output['stdout']:
                                            prerequisitesSatisfied = False
                                            break

                                    if prerequisitesSatisfied and not prompt:
                                        return promptRHCert(key)

                        # Else, determine if the child has written any new data to stderr.
                        elif selectEvent[0] == child.stderr.fileno() and selectEvent[1] == select.POLLIN:
                            # child.stderr.flush()
                            # Read the child's output on stderr,
                            # store the information locally,
                            # and log it appropriately.
                            output = child.stderr.read()
                            print(output.strip())
                            if usePY3:
                                child_output['stderr'] = child_output['stderr'] + output.decode()
                            else:
                                child_output['stderr'] = child_output['stderr'] + output

                        # Else, check to see if the select event
                        # has an event code equal to select.POLLHUP.
                        # This indicates that the child has likely terminated properly.
                        # elif selectEvent[1] loop.
                        elif (selectEvent[1] >= select.POLLHUP and selectEvent[1] < select.POLLNVAL) or selectEvent[1] >= (select.POLLHUP + select.NVAL):
                            channelHUP.append(selectEvent[0])

                    # If select() returns a select.POLLHUP event for all
                    # registered channels, we can safely assume that the process
                    # terminated successfully. We can now return the
                    # returncode, stdout, and stderr of the rhcert command
                    # by breaking the while loop.
                    if len(channelHUP) == len(pollData) and len(pollData) != 0:
                        return

            promptRHCert()

            # Although the process has already terminated, we need to
            # explicitly call communicate() to ensure the exist status is properly set.
            child.communicate()

            return (child.returncode, child_output['stdout'], child_output['stderr'])

        # Else, the command specified by the calling function is not
        # associated with a redhat-certification utility.
        # We can go ahead and run this command normally.
        (child_output['stdout'], child_output['stderr']) = child.communicate()

        return (child.returncode, child_output['stdout'], child_output['stderr'])

    except Exception as e:
        printError(e, raiseError=True)

#@given('I visit {url}')
#def step_impl(context, url):
#    context.browser.get(url)
#
#
#@when(u'I select \'{value}\' from the \'{select}\' dropdown')
#def step_impl(context, value, select):
#    chosen = context.browser.find_element_by_id(HTML_LOOKUPS[select])
#    chosen.find_element_by_css_selector("a").click()
#    results = chosen.find_elements_by_css_selector(".chosen-results li")
#    for result in results:
#        if result.text == value:
#            break
#    result.click()
#
#
#@when('I fill in \'{textbox}\' with \'{text}\'')
#def step_impl(context, textbox, text):
#    selected_textbox = context.browser.find_element_by_id(HTML_LOOKUPS[textbox])
#    selected_textbox.clear()
#    selected_textbox.send_keys(text)
#
#
#@when('I click on \'{button}\'')
#def step_impl(context, button):
#    selected_button = context.browser.find_element_by_id(HTML_LOOKUPS[button])
#    selected_button.click()
#
#def safe_selenium_imports(filename, scope):
#    def raiseImportError():
#        raise ImportError("Unable to import Selenium on unsupported Python 2.6 version from file {0}.".format(filename))
#
#    try:
#        from selenium import webdriver
#        from selenium import common
#        from selenium.webdriver.support import expected_conditions as ec
#        scope['webdriver'] = webdriver
#        scope['common'] = common
#        scope['webdriver'].support.expected_conditions = ec
#    except SyntaxError as e:
#        print("A SyntaxError was observed while importing")
#        print("the python-selenium module within {0}.".format(__file__))
#        print("{0}: {1}".format(e.__class__, e))
#        print("This only occurs for EL6 (python-2.6), and is therefore")
#        print("only a true error if we are assigned a 'SERVERS' role in Beaker.")
#
#        # To maintain default behavior, we're going to go ahead and reraise
#        # this exception if Spicegate is not running on Beaker systems
#        # as part of our CI workflow.
#        if "SERVERS" not in os.environ and "CLIENTS" not in os.environ:
#            raiseImportError()
#
#        # Else, reraise an exception if the system running Spicegate is
#        # fulfilling a "SERVERS" role within a multi-host testing environment
#        # within our CI workflow.
#        if os.environ["SERVERS"].find(socket.gethostname()) != -1:
#            raiseImportError()
#
#    except ImportError as e:
#        print("A {0} was encountered while importing selenium from {1}.".format(e.__class__, filename))
#        print("{0}: {1}".format(e.__class__, e))
#        raise e
